# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_callback.ipynb (unless otherwise specified).

__all__ = ['GatherInputsCallback', 'SetInputsCallback', 'GeneratorCallback']

# Cell
from fastcore.basics import store_attr
from fastcore.meta import delegates

from fastai.callback.core import Callback, CancelBatchException

from transformers import PreTrainedModel

# Cell
class GatherInputsCallback(Callback):
    """
    Prepares basic input dictionary for HuggingFace Transformers

    This `Callback` generates a very basic dictionary consisting of `input_ids`,
    `attention_masks`, and `token_type_ids`, and saves it to the attribute `self.learn.inputs`.

    If further data is expected or needed from the batch, the additional Callback(s) should have
    an order of -2
    """
    order = -3

    def before_validate(self):
        """
        Sets the number of inputs in `self.dls`
        """
        x = self.dl.one_batch()
        self.learn.dls.n_inp = len(x)

    def before_batch(self):
        """
        Turns `self.xb` from a tuple to a dictionary of either
            `{"input_ids", "attention_masks", "token_type_ids"}`d
        or
            `{"input_ids", "attention_masks"}`
        """
        inputs = {
                "input_ids":self.learn.xb[0],
                "attention_mask":self.learn.xb[1]
        }

        if len(self.learn.xb) > 2:
            inputs["token_type_ids"] = self.learn.xb[2]

        self.learn.inputs = inputs

# Cell
class SetInputsCallback(Callback):
    """
    Callback which runs after `GatherInputsCallback` that sets `self.learn.xb`
    """
    order = -1

    def __init__(
        self,
        as_dict=False # Whether to leave `self.xb` as a dictionary of values
    ): store_attr()

    def before_batch(self):
        """
        Set `self.learn.xb` to `self.learn.inputs.values()`
        """
        if not self.as_dict:
            self.learn.xb = list(self.learn.inputs.values())
        else:
            self.learn.xb = self.learn.inputs

# Cell
class GeneratorCallback(Callback):
    """
    Callback used for models that utilize `self.model.generate`
    """

    @delegates(PreTrainedModel.generate)
    def __init__(
        self,
        num_beams:int, # Number of beams for beam search
        min_length:int, # Minimal length of the sequence generated
        max_length:int, # Maximum length of the sequence generated
        early_stopping:bool, # Whether to do early stopping
        **kwargs
    ):
        store_attr()
        self.kwargs = kwargs

    def before_batch(self):
        """
        Run model-specific inference
        """

        pred = self.learn.model.generate(
            input_ids = self.xb['input_ids'],
            attention_mask = self.xb['attention_mask'],
            num_beams = self.num_beams,
            min_length = self.min_length,
            max_length = self.max_length,
            early_stopping = self.early_stopping,
            **self.kwargs
        )

        self.learn.pred = pred

        raise CancelBatchException # skip original model inference


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_sequence_classification.ipynb (unless otherwise specified).

__all__ = ['logger', 'SequenceResult', 'TransformersSequenceClassifier', 'FlairSequenceClassifier',
           'EasySequenceClassifier']

# Cell
import logging
from typing import List, Dict, Union, Tuple, Callable
from collections import defaultdict, OrderedDict
from pathlib import Path

import torch
from torch import nn
from torch.utils.data import TensorDataset, DataLoader
import datasets
from datasets import ClassLabel
from sklearn.metrics import accuracy_score, precision_recall_fscore_support
from flair.data import Sentence, DataPoint
from flair.models import TextClassifier
from transformers import (
    AutoTokenizer,
    AutoModelForSequenceClassification,
    PreTrainedTokenizer,
    PreTrainedModel,
    BertPreTrainedModel,
    DistilBertPreTrainedModel,
    XLMPreTrainedModel,
    XLNetPreTrainedModel,
    ElectraPreTrainedModel,
    BertForSequenceClassification,
    XLNetForSequenceClassification,
    AlbertForSequenceClassification,
    TrainingArguments,
    Trainer,
)

from ..model import AdaptiveModel
from ..model_hub import HFModelResult, FlairModelResult

from fastcore.basics import risinstance
from fastcore.xtras import Path
from ..result import DetailLevel, SentenceResult

from torch import tensor

# Cell
logger = logging.getLogger(__name__)

# Cell
class SequenceResult(SentenceResult):
    "A result class designed for Sequence Classification models"
    def __init__(
        self,
        sentences:List[Sentence], # A list of flair `Sentence`'s
        class_names:list = None # A potential list of class names
    ):
        super().__init__(sentences)
        self.classes = sentences[0].get_label_names()
        self.class_names = class_names

    @property
    def probabilities(self) -> List[List[tensor]]:
        """
        The probabilities returned for each classification
        """
        return torch.stack([tensor(list(map(lambda x: x.score, i.get_labels()))) for i in self._sentences], dim=0)

    @property
    def predictions(self) -> List[str]:
        """
        A list of the best classification for each input
        """
        if self.class_names is not None:
            return [self.class_names[p.argmax()] for p in self.probabilities]
        return [max(s.labels, key=lambda x: x.score).value for s in self._sentences]

    def to_dict(
        self,
        detail_level:DetailLevel=DetailLevel.Low # A level of detail to return
    ) -> dict:
        "Return `self` as a dictionary"
        o = {
            'sentences':self.inputs,
            'predictions':self.predictions,
            'probs':self.probabilities
        }
        if detail_level == 'medium' or detail_level == 'high':
            # Add a dictionary of sentence and probabilities, and return the vocab
            o['pairings'] = OrderedDict({
                s:probs for (s,probs) in zip(self.inputs, self.probabilities)
            })
            o['classes'] = self.classes

        if detail_level == 'high':
            # Add original `Sentences`
            o['sentences'] = self._sentences
        return o

# Cell
class TransformersSequenceClassifier(AdaptiveModel):
    "Adaptive model for Transformer's Sequence Classification Model"

    def __init__(
        self,
        tokenizer: PreTrainedTokenizer, # A tokenizer object from Huggingface's transformers (TODO)and tokenizers
        model: PreTrainedModel # A transformers Sequence Classification model
    ):
        # Load up model and tokenizer
        self.tokenizer = tokenizer
        super().__init__()
        self.set_model(model)

        # Load empty trainer
        self.trainer = None

        # Setup cuda and automatic allocation of model
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.model.to(self.device)

    @classmethod
    def load(
        cls,
        model_name_or_path: Union[HFModelResult, str] # A key string of one of Transformer's pre-trained Sequence Classifier Model or a `HFModelResult`
    ) -> AdaptiveModel:
        "Class method for loading and constructing this classifier"
        if isinstance(model_name_or_path, HFModelResult): model_name_or_path = model_name_or_path.name
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, use_fast=True)
        model = AutoModelForSequenceClassification.from_pretrained(model_name_or_path)
        classifier = cls(tokenizer, model)
        return classifier

    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Sentences to run inference on
        mini_batch_size: int = 32, # Mini batch size
        **kwargs, # Optional arguments for the Transformers classifier
    ) -> List[Sentence]: # Returns a list of `Sentence` predictions
        "Predict method for running inference using the pre-trained sequence classifier model"
        id2label = self.model.config.id2label
        sentences = text
        results: List[Sentence] = []

        if not sentences: return sentences

        if risinstance([DataPoint, str], sentences):
            sentences = [sentences]

        # filter empty sentences
        if isinstance(sentences[0], Sentence):
            sentences = [sentence for sentence in sentences if len(sentence) > 0]
        if len(sentences) == 0:
            return sentences

        # reverse sort all sequences by their length
        rev_order_len_index = sorted(
            range(len(sentences)), key=lambda k: len(sentences[k]), reverse=True
        )
        original_order_index = sorted(
            range(len(rev_order_len_index)), key=lambda k: rev_order_len_index[k]
        )

        reordered_sentences: List[Union[DataPoint, str]] = [
            sentences[index] for index in rev_order_len_index
        ]

        # Turn all Sentence objects into strings
        if isinstance(reordered_sentences[0], Sentence):
            str_reordered_sentences = [
                sentence.to_original_text() for sentence in sentences
            ]
        else:
            str_reordered_sentences = reordered_sentences

        dataset = self._tokenize(str_reordered_sentences)
        dl = DataLoader(dataset, batch_size=mini_batch_size)
        predictions: List[Tuple[str, float]] = []

        outputs, _ = super().get_preds(dl=dl)
        logits = torch.cat([o['logits'] for o in outputs])
        predictions = torch.softmax(logits, dim=1).tolist()

        for text, pred in zip(str_reordered_sentences, predictions):
            # Initialize and assign labels to each class in each datapoint prediction
            text_sent = Sentence(text)
            for k, v in id2label.items():
                text_sent.add_label(typename='sc', value=v, score=pred[k])
            results.append(text_sent)

        # Order results back into original order
        results = [results[index] for index in original_order_index]

        return results

    def _tokenize(
        self, sentences: Union[List[Sentence], Sentence, List[str], str]
    ) -> TensorDataset:
        """ Batch tokenizes text and produces a `TensorDataset` with them """

        # TODO: __call__ from tokenizer base class in the transformers library could automate/handle this
        tokenized_text = self.tokenizer.batch_encode_plus(
            sentences,
            return_tensors='pt',
            max_length=None,
            add_special_tokens=True,
        )

        # Bart, XLM, DistilBERT, RoBERTa, and XLM-RoBERTa don't use token_type_ids
        if isinstance(
            self.model,
            (
                BertForSequenceClassification,
                XLNetForSequenceClassification,
                AlbertForSequenceClassification,
            ),
        ):
            dataset = TensorDataset(
                tokenized_text['input_ids'],
                tokenized_text['attention_mask'],
                tokenized_text['token_type_ids'],
            )
        else:
            dataset = TensorDataset(
                tokenized_text['input_ids'], tokenized_text['attention_mask']
            )

        return dataset

# Cell
class FlairSequenceClassifier(AdaptiveModel):
    "Adaptive Model for Flair's Sequence Classifier"

    def __init__(
        self,
        model_name_or_path: str # A key string of one of Flair's pre-trained Sequence Classifier Model
    ):
        self.classifier = TextClassifier.load(model_name_or_path)

    @classmethod
    def load(
        cls,
        model_name_or_path: Union[HFModelResult, FlairModelResult, str] # A key string of one of Flair's pre-trained Sequence Classifier Model or a `HFModelResult`
    ) -> AdaptiveModel:
        "Class method for loading a constructing this classifier"
        if risinstance([HFModelResult, FlairModelResult], model_name_or_path):
            model_name_or_path = model_name_or_path.name
        classifier = cls(model_name_or_path)
        return classifier

    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Sentences to run inference on
        mini_batch_size: int = 32, # Mini batch size
        **kwargs, # Optional arguments for the Flair classifier
    ) -> List[Sentence]: # A list of predicted `Sentence`s
        "Predict method for running inference using the pre-trained sequence classifier model"
        if isinstance(text, (Sentence, str)):
            text = [text]
        if isinstance(text[0], str):
            text = [Sentence(s) for s in text]

        self.classifier.predict(
            sentences=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )

        return text

# Cell
from ..model_hub import HFModelHub, FlairModelHub

# Cell
class EasySequenceClassifier:
    "Easy module for sequence classifiers"

    def __init__(self):
        self.sequence_classifiers: Dict[AdaptiveModel] = defaultdict(bool)
        self.hf_hub = HFModelHub()
        self.flair_hub = FlairModelHub()

    def tag_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # String, list of strings, `Sentence`, or list of `Sentence`s to be classified
        model_name_or_path: Union[str, FlairModelResult, HFModelResult] = 'en-sentiment', # The model name key or model path
        mini_batch_size: int = 32, # The mini batch size for running inference
        detail_level:DetailLevel = DetailLevel.Low, # A level of detail to return
        class_names:list = None, # A list of labels
        **kwargs, # Keyword Arguments for Flair's `TextClassifier.predict()` method params
    ) -> List[Sentence]: # A list of Flair's `Sentence`'s
        "Tags a text sequence with labels the sequence classification models have been trained on"
        # Load Text Classifier Model and Pytorch Module into tagger dict
        name = getattr(model_name_or_path, 'name', model_name_or_path)
        if not self.sequence_classifiers[name]:
            """
            self.sequence_classifiers[name] = TextClassifier.load(
                model_name_or_path
            )
            """
            if risinstance([FlairModelResult, HFModelResult], model_name_or_path):
                try:
                    self.sequence_classifiers[name] = FlairSequenceClassifier.load(name)
                except:
                    self.sequence_classifiers[name] = TransformersSequenceClassifier.load(name)

            elif risinstance([str, Path], model_name_or_path) and (Path(model_name_or_path).exists() and Path(model_name_or_path).is_dir()):
                # Load in previously existing model
                try:
                    self.sequence_classifiers[name] = FlairSequenceClassifier.load(name)
                except:
                    self.sequence_classifiers[name] = TransformersSequenceClassifier.load(name)

            else:
                # Flair
                res = self.flair_hub.search_model_by_name(name, user_uploaded=True)
                if len(res) < 1:
                    # No models found
                    res = self.hf_hub.search_model_by_name(model_name_or_path, user_uploaded=True)
                    if len(res) < 1:
                        logger.info("Not a valid `model_name_or_path` param")
                        return [Sentence('')]
                    else:
                        name = res[0].name.replace('flairNLP', 'flair')
                        self.sequence_classifiers[res[0].name] = TransformersSequenceClassifier.load(name)
                else:
                    name = res[0].name.replace('flairNLP/', '')
                    self.sequence_classifiers[name] = FlairSequenceClassifier.load(name) # Returning the first should always be non-fast

        classifier = self.sequence_classifiers[name]
        out = classifier.predict(
            text=text,
            mini_batch_size=mini_batch_size,
            **kwargs,
        )
        if detail_level is None: return out
        res = SequenceResult(out, class_names)
        return res.to_dict(detail_level)


    def tag_all(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        mini_batch_size: int = 32, # The mini batch size for running inference
        detail_level:DetailLevel = DetailLevel.Low, # A level of detail to return
        class_names:list = None, # A list of labels
        **kwargs, # Keyword Arguments for Flair's `TextClassifier.predict()` method params
    ) -> List[Sentence]: # A list of Flair's `Sentence`'s
        "Tags text with all labels from all sequence classification models"
        sentences = text
        for tagger_name in self.sequence_classifiers.keys():
            sentences = self.tag_text(
                sentences,
                model_name_or_path=tagger_name,
                mini_batch_size=mini_batch_size,
                detail_level = None
                **kwargs,
            )
        res = SequenceResult(out, class_names)
        return res.to_dict(detail_level)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_summarization.ipynb (unless otherwise specified).

__all__ = ['logger', 'TransformersSummarizer', 'EasySummarizer']

# Cell
import logging
from typing import List, Dict, Union
from collections import defaultdict

import torch
from torch.utils.data import TensorDataset, DataLoader

from transformers import (
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
    PreTrainedTokenizer,
    PreTrainedModel,
    T5ForConditionalGeneration,
    BartForConditionalGeneration,
)

from ..callback import GeneratorCallback
from ..model import AdaptiveModel
from ..model_hub import HFModelResult, FlairModelResult

from fastcore.basics import store_attr
from fastcore.meta import delegates

from fastai.callback.core import Callback, CancelBatchException
from fastai.torch_core import apply

# Cell
logger = logging.getLogger(__name__)

# Cell
class TransformersSummarizer(AdaptiveModel):
    "Adaptive model for Transformer's Conditional Generation or Language Models (Transformer's T5 and Bart conditiional generation models have a language modeling head)"
    def __init__(
        self,
        tokenizer: PreTrainedTokenizer, # A tokenizer object from Huggingface's transformers (TODO)and tokenizers
        model: PreTrainedModel # A transformers Conditional Generation (Bart or T5) or Language model
    ):
        # Load up tokenizer
        self.tokenizer = tokenizer

        super().__init__()
        # Sets internal model
        super().set_model(model)

        # Set inputs to come in as `dict`
        super().set_as_dict(True)

    @classmethod
    def load(
        cls,
        model_name_or_path: str # A key string of one of Transformer's pre-trained Summarizer Model
    ) -> AdaptiveModel:
        "Class method for loading and constructing this classifier"
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_name_or_path)
        summarizer = cls(tokenizer, model)
        return summarizer

    def predict(
        self,
        text: Union[List[str], str], # Sentences to run inference on
        mini_batch_size: int = 32, # Mini batch size
        num_beams: int = 4, # Number of beams for beam search. Must be between 1 and infinity. 1 means no beam search.
        min_length: int = 0, # The min length of the sequence to be generated
        max_length: int = 128, # The max length of the sequence to be generated. Between min_length and infinity
        early_stopping: bool = True, # If set to True beam search is stopped when at least num_beams sentences finished per batch
        **kwargs, # Optional arguments for the Transformers `PreTrainedModel.generate()` method
    ) -> List[str]: # A list of predicted summarizations
        "Predict method for running inference using the pre-trained sequence classifier model"

        # Make all inputs list
        if isinstance(text, str):
            text = [text]

        # T5 requires 'summarize: ' precursor text for pre-trained summarizer
        if isinstance(self.model, T5ForConditionalGeneration):
            text = [f'summarize: {t}' for t in text]

        dataset = self._tokenize(text)
        dl = DataLoader(dataset, batch_size=mini_batch_size)

        summaries = []

        logger.info(f'Running summarizer on {len(dataset)} text sequences')
        logger.info(f'Batch size = {mini_batch_size}')

        cb = GeneratorCallback(num_beams, min_length, max_length, early_stopping, **kwargs)

        preds,_ = super().get_preds(dl=dl, cbs=[cb])

        preds = apply(lambda x: x.squeeze(0), preds)

        for o in preds:
            summaries.append(
                [
                    self.tokenizer.decode(
                        o,
                        skip_special_tokens=True,
                        clean_up_tokenization_spaces=False,
                    )
                ].pop()
            )

        return {'summaries':summaries}

    def _tokenize(self, text: Union[List[str], str]) -> TensorDataset:
        "Batch tokenizes text and produces a `TensorDataset` with text"

        # Pre-trained Bart summarization model has a max length fo 1024 tokens for input
        if isinstance(self.model, BartForConditionalGeneration):
            tokenized_text = self.tokenizer.batch_encode_plus(
                text,
                return_tensors="pt",
                max_length=1024,
                add_special_tokens=True,
                padding="max_length",
            )
        else:
            tokenized_text = self.tokenizer.batch_encode_plus(
                text,
                return_tensors="pt",
                padding="max_length",
                add_special_tokens=True,
            )

        # Bart doesn't use `token_type_ids`
        dataset = TensorDataset(
            tokenized_text["input_ids"],
            tokenized_text["attention_mask"],
        )

        return dataset

# Cell
class EasySummarizer:
    "Summarization Module"
    def __init__(self):
        self.summarizers: Dict[AdaptiveModel] = defaultdict(bool)

    def summarize(
        self,
        text: Union[List[str], str], # Sentences to run inference on
        model_name_or_path: Union[str, HFModelResult] = "t5-small", # A model id or path to a pre-trained model repository or custom trained model directory
        mini_batch_size: int = 32, # Mini batch size
        num_beams: int = 4, # Number of beams for beam search. Must be between 1 and infinity. 1 means no beam search
        min_length: int = 0, # The max length of the sequence to be generated. Between min_length and infinity
        max_length: int = 128, # The max length of the sequence to be generated. Between min_length and infinity
        early_stopping: bool = True, # If set to True beam search is stopped when at least num_beams sentences finished per batch
        **kwargs, # Optional arguments for the Transformers `PreTrainedModel.generate()` method
    ) -> List[str]: # A list of predicted summaries
        "Predict method for running inference using the pre-trained sequence classifier model"
        name = getattr(model_name_or_path, 'name', model_name_or_path)
        if not self.summarizers[name]:
            self.summarizers[name] = TransformersSummarizer.load(
                name
            )

        summarizer = self.summarizers[name]
        return summarizer.predict(
            text=text,
            mini_batch_size=mini_batch_size,
            num_beams=num_beams,
            min_length=min_length,
            max_length=max_length,
            early_stopping=early_stopping,
            **kwargs,
        )


__version__ = "0.3.0"

from pathlib import Path

# global variable like Flair's cache_root
cache_root = Path.home()/".adaptnlp"

# Inference modules
from .inference.embeddings import (
    EasyWordEmbeddings,
    EasyStackedEmbeddings,
    EasyDocumentEmbeddings
)

from .inference.sequence_classification import (
    EasySequenceClassifier,
    TransformersSequenceClassifier,
    FlairSequenceClassifier
)

from .inference.token_classification import EasyTokenTagger
from .inference.question_answering import EasyQuestionAnswering, TransformersQuestionAnswering
from .inference.summarization import EasySummarizer, TransformersSummarizer
from .inference.translation import EasyTranslator, TransformersTranslator
from .inference.text_generation import EasyTextGenerator, TransformersTextGenerator

from .result import DetailLevel

# Huggingface Hub bits
from .model_hub import HFModelHub, FlairModelHub, HF_TASKS, FLAIR_TASKS

# Training API
from .training.core import Strategy, TaskDatasets, AdaptiveTuner
from .training.sequence_classification import SequenceClassificationTuner, SequenceClassificationDatasets
from .training.language_model import LanguageModelTuner, LanguageModelDatasets


__all__ = [
    "__version__",
    "EasyWordEmbeddings",
    "EasyStackedEmbeddings",
    "EasyDocumentEmbeddings",
    "EasySequenceClassifier",
    "TransformersSequenceClassifier",
    "FlairSequenceClassifier",
    "EasyTokenTagger",
    "EasyQuestionAnswering",
    "TransformersQuestionAnswering",
    "EasySummarizer",
    "TransformersSummarizer",
    "EasyTranslator",
    "TransformersTranslator",
    "EasyTextGenerator",
    "TransformersTextGenerator",
    "DetailLevel",
    "HFModelHub",
    "FlairModelHub",
    "HF_TASKS",
    "FLAIR_TASKS",
    "TaskDatasets",
    "AdaptiveTuner",
    "Strategy",
    "SequenceClassificationDatasets",
    "LanguageModelDatasets",
    "SequenceClassificationTuner",
    "LanguageModelTuner"
]


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/12_training.core.ipynb (unless otherwise specified).

__all__ = ['Strategy', 'ParentLabeller', 'ColReader', 'Categorize', 'MultiCategorize', 'RandomSplitter', 'TaskDatasets',
           'AdaptiveDataLoaders', 'AdaptiveTuner']

# Cell
from fastcore.xtras import Path, range_of # pathlib `Path` with extra bits
from fastcore.foundation import mask2idxs, L
from fastcore.meta import delegates
from fastcore.basics import mk_class

from fastai.learner import Learner
from fastai.callback.hook import Learner
from fastai.callback.progress import Learner
from fastai.callback.schedule import Learner
# Patch'd Learner functionalities

from fastai.data.core import DataLoaders

from torch.utils.data import DataLoader

from transformers import default_data_collator, AutoTokenizer, AutoModel
import torch

from typing import List, Union

import pandas as pd

# Cell
#nbdev_comment _all_ = ['Strategy']

# Cell
class ParentLabeller:
    """
    Extracts class based on filename's parent at `level`
    """
    def __init__(
        self,
        level=1 # The level up from `fname` to find the label
    ):
        self.level = level

    def __call__(self, o:Path): return self._do_level(o, self.level)

    def _do_level(self, o:Path, level:int):
        "Goes down one level on parent"
        def _inner(a): return a.parent
        if level == 1: return o.parent.name
        else: return self._do_level(_inner(o), level - 1)

# Cell
class ColReader:
    """
    Reads `cols` in `row` with potential `pref` and `suff`
    Based on the fastai class
    """
    def __init__(
        self,
        cols, # Some column names to use
        pref:str='', # A prefix
        suff:str='', # A suffix
        label_delim:str=None, # A label delimiter
    ):
        self.pref = str(pref) + os.path.sep if isinstance(pref, Path) else pref
        self.suff, self.label_delim = suff, label_delim
        self.cols = L(cols)

    def _do_one(self, r, c):
        o = r[c] if isinstance(c,int) else r[c] if c=='name' or c=='cat' else getattr(r,c)
        if len(self.pref)==0 and len(self.suff)==0 and self.label_delim is None: return o
        if self.label_delim is None: return f'{self.pref}{o}{self.suff}'
        else: return o.split(self.label_delim) if len(o)>0 else []

    def __call__(self, o):
        if len(self.cols) == 1: return self._do_one(o, self.cols[0])
        return L(self._do_one(o,c) for c in self.cols)

# Cell
class Categorize:
    """
    Collection of categories with reverse mapping in `o2i`
    Based on the fastai class
    """
    def __init__(
        self,
        names, # An interable collection of items to create a vocab from
        sort=True # Whether to make the items sorted
    ):
        names = L(names)
        self.classes = L(o for o in names.unique() if o == o)
        if sort: self.classes = self.classes.sorted()
        self.o2i = dict(self.classes.val2idx())

    def map_objs(
        self,
        objs # Some iterable collection
    ) -> L:
        "Map `objs` to IDs"
        return L(self.o2i[o] for o in objs)

    def map_ids(
        self,
        ids # Some ids correlating to `self.classes`
    ) -> L:
        "Map `ids` to objects in vocab"
        return L(self.classes[o] for o in ids)

    def __call__(self, o):
        "Label encode a single `o`"
        return int(self.o2i[o])

    def decode(
        self,
        o # A key in self.classes
    ):
        "Decodes `o` by looking in `self.classes`"
        return self.classes[o]

# Cell
class MultiCategorize(Categorize):
    """
    Collection of multi-categories with reverse mapping in `o2i`
    Based on the fastai class
    """
    def __init__(
        self,
        names, # An interable collection of items to create a vocab from
    ):
        super().__init__(names=names, sort=names==None)

    def __call__(self, o):
        "Label encode a single `o`"
        if not all(nm in self.o2i.keys() for nm in o):
            diff = [e for e in o if e not in self.o2i.keys()]
            diff_str = "', '".join(diff)
            raise KeyError(f"Labels '{diff_str}' were not included in the training dataset")

        return [int(self.o2i[o_]) for o_ in o]

    def decode(
        self,
        o # A list of keys in self.classes
    ) -> list:
        "Decodes `o` by looking in `self.classes`"
        return [self.classes[o_] for o_ in o]

# Cell
def RandomSplitter(valid_pct=0.2, seed=None):
    """
    Creates a function that splits some items between train and validation with `valid_pct` randomly
    Based on the fastai class
    """
    def _inner(o):
        if seed is not None: torch.manual_seed(seed)
        rand_idx = L(list(torch.randperm(len(o)).numpy()))
        cut = int(valid_pct * len(o))
        return rand_idx[cut:], rand_idx[:cut]
    return _inner

# Internal Cell
def _base_tok(item, tokenizer, tokenize_kwargs): return tokenizer(item, **tokenize_kwargs)

# Cell
class TaskDatasets:
    """
    A set of datasets for a particular task, with a simple API.

    Note: This is the base API, `items` should be a set of regular text and model-ready labels,
          including label or one-hot encoding being applied.
    """
    def __init__(
        self,
        train_dset, # A train `Dataset` object
        valid_dset, # A validation `Dataset` object
        tokenizer_name:str = None, # The string name of a `HuggingFace` tokenizer or model. If `None`, will not tokenize the dataset.
        tokenize:bool = True, # Whether to tokenize the dataset immediatly
        tokenize_func:callable = None, # A function to tokenize an item with
        tokenize_kwargs:dict = {}, # Some kwargs for when we call the tokenizer
        auto_kwargs:dict = {}, # Some kwargs when calling `AutoTokenizer.from_pretrained`
        remove_cols:Union[str,List[str]] = None, # What columns to remove
    ):
        self.train = train_dset
        self.valid = valid_dset
        self.tokenizer = None
        self.remove_cols = listify(remove_cols)
        if tokenizer_name is not None: self.set_tokenizer(tokenizer_name, **auto_kwargs)
        if tokenize_func is not None: self.tok_func = tokenize_func
        else: self.tok_func = _base_tok
        if self.tokenizer:
            if 'max_length' in tokenize_kwargs.keys() and self.tokenizer.model_max_length >= tokenize_kwargs['max_length']: pass
            elif 'max_length' in tokenize_kwargs.keys() and self.tokenizer.model_max_length < tokenize_kwargs['max_length']:
                print("Warning: `max_length` is larger than the pretrained model")
            elif 'max_length' not in tokenize_kwargs.keys():
                print("No value for `max_length` set, automatically adjusting to the size of the model and including truncation")
                tokenize_kwargs['max_length'] = self.tokenizer.model_max_length
                tokenize_kwargs['truncation'] = True
                print(f"Sequence length set to: {tokenize_kwargs['max_length']}")
        self.tokenize_kwargs = tokenize_kwargs
        if tokenize and self.tokenizer is not None: self._tokenize()
        elif tokenize and self.tokenizer is None:
            print("Tried to tokenize a dataset without a tokenizer. Please set a tokenizer with `set_tokenizer` and call `_tokenize()`")


    def __getitem__(self, idx): return self.train[idx]

    def _tokenize(self):
        "Tokenize dataset in `self.items` with `kwargs` for `tokenize()`"
        if not self.tokenizer: raise ValueError("Tried to tokenize a dataset without a tokenizer. Please add a tokenizer with `set_tokenizer(tokenizer_name` and try again")
        f = partial(self.tok_func, tokenizer=self.tokenizer, tokenize_kwargs=self.tokenize_kwargs)
        self.train = self.train.map(f,batched=True,remove_columns = self.remove_cols)
        self.valid = self.valid.map(f,batched=True,remove_columns = self.remove_cols)

    @delegates(AutoTokenizer.from_pretrained)
    def set_tokenizer(
        self,
        tokenizer_name:str, # A string name of a `HuggingFace` tokenizer or model
        override_existing:bool = False, # Whether to override an existing tokenizer
        **kwargs # kwargs to go to `AutoTokenizer.from_pretrained`
    ):
        "Sets a new `AutoTokenizer` to `self.tokenizer`"
        if self.tokenizer and not override_existing:
            print(f'Warning! You are trying to override an existing tokenizer: {self.tokenizer.name_or_path}. Pass `override_existing=True` to use a new tokenizer')
            return
        elif self.tokenizer and override_existing:
            print(f'Setting new tokenizer to {tokenizer_name}')
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name, **kwargs)
        except:
            raise ValueError(f'{tokenizer_name} is not a valid pretrained model on the HuggingFace Hub or a local model')

    def set_classes(
        self,
        classes:list, # An ordered list of class names
    ):
        "Override the class labels in `self.categorize` it exists, otherwise make it"
        if hasasttr(self, 'categorize'):
            self.categorize.classes = L(classes).sorted()
            self.categorize.o2i = dict(self.categorize.val2idx())
        else:
            self.categorize = Categorize(classes)


    @delegates(DataLoaders)
    def dataloaders(
        self,
        batch_size:int=8, # A batch size
        shuffle_train:bool=True, # Whether to shuffle the training dataset
        collate_fn:callable = None, # A custom collation function
        **kwargs # Torch DataLoader kwargs
    ):
        "Creates `DataLoaders` from the dataset"
        if collate_fn is None: collate_fn = default_data_collator
        train_dl = DataLoader(self.train, shuffle=shuffle_train, collate_fn=collate_fn, batch_size=batch_size, **kwargs)
        valid_dl = DataLoader(self.valid, shuffle=False, collate_fn=collate_fn, batch_size=batch_size, **kwargs)
        return AdaptiveDataLoaders(train_dl, valid_dl, tokenizer=self.tokenizer)

# Cell
class AdaptiveDataLoaders(DataLoaders):
    "A set of `DataLoaders` that keeps track of a `tokenizer`"
    def __init__(
        self,
        *loaders, # A variable list of DataLoaders
        tokenizer=None, # A Transformers tokenizer object
        path='.', # A path to be stored in `self.path`
        device=None # A device for the tensors, such as "cpu" or "cuda:0"
    ):
        self.tokenizer = tokenizer
        super().__init__(*loaders, path=path, device=device)

    def show_batch(
        self,
        ds_idx:int=0, # Index of the DataLoader to show, 0 = training, 1 = validation
        n:int=5, # Number of examples to show
        raw:bool=False, # Prints the raw inputs and targets without decoding
    ):
        "Show a batch of data"
        dl = self[ds_idx]
        batch = next(iter(self[ds_idx]))

        if n > len(batch['input_ids']):
            print('`n` is larger than one batch, printing entire batch')
            n = len(batch['input_ids'])
        if n < 1:
            raise ValueError('Tried to show zero samples, please enter a value for `n` greater than 0')

        if not self.tokenizer and not raw:
            print("Cannot decode without a tokenizer, printing raw outputs..")

        if raw or not self.tokenizer:
            new_cols = []
            for nm in batch.keys():
                new_cols.append(nm)
                new_cols.append(f'{nm} Shape')
            df = pd.DataFrame(columns=new_cols)
            for i in range(n):
                new_row = []
                for key in batch.keys():
                    new_row += [batch[key][i].cpu().numpy()]
                    new_row += [batch[key][i].shape]
                df.loc[i] = new_row

        else:
            inputs = self.tokenizer.batch_decode(
                batch['input_ids'],
                skip_special_tokens=True
            )
            lbls = batch['labels']

            df = pd.DataFrame(columns=['Input', 'Label'])
            if hasattr(self, 'categorize'):
                lbls = [self.categorize.decode(o.cpu().numpy()) for o in lbls]
            elif len(lbls.shape) < len(batch['input_ids'].shape):
                # Not a language model, but we can't decode
                lbls = lbls
            else:
                # It's a language model
                lbls = self.tokenizer.batch_decode(lbls, skip_special_tokens=True)
            for i in range(n):
                df.loc[i] = [inputs[i], lbls[i]]
        display_df(df)

# Cell
class _AdaptiveLearner(Learner):
    """
    A base fastai `Learner` that overrides `_split` and `_do_one_batch` to
    have it work with HuggingFace datasets and models
    """
    def _split(self, b):
        "Assign `self.xb` to model input and labels"
        self.xb = b
        if 'labels' in b.keys(): self.yb = b['labels'].unsqueeze(0)

    def _do_one_batch(self):
        "Move a batch of data to a device, get predictions, calculate the loss, and perform backward pass"
        self.xb = {k:v.to(self.device) for k,v in self.xb.items()} # See if `to_device` fixes this
        self.yb = self.yb.to(self.device)
        out = self.model(**self.xb)
        self.pred = out['logits'].to(self.device)
        self('after_pred')
        self.loss_grad = out['loss'].to(self.device)
        self.loss = self.loss_grad.clone()
        self('after_loss')
        if not self.training or not len(self.yb): return
        self('before_backward')
        self.loss_grad.backward()
        self._with_events(self.opt.step, 'step', CancelStepException)
        self.opt.zero_grad()

# Cell
mk_class('Strategy', **{'OneCycle':'fit_one_cycle', 'CosineAnnealing':'fit_flat_cos', 'SGDR':'fit_sgdr'}, doc='Class for fitting strategies with typo-proofing')

# Cell
class AdaptiveTuner:
    "A base `Tuner` that interfaces with `AdaptiveLearner` with specific exposed functions"
    def __init__(
        self,
        expose_fastai:bool=False, # Whether to expose the entire API in `self`
        tokenizer = None, # A HuggingFace tokenizer
        **kwargs # kwargs for `_AdaptiveLearner`
    ):
        self.tokenizer = tokenizer
        self._tuner = _AdaptiveLearner(**kwargs)

        exposed_attrs = ['dls', 'model', 'loss_func', 'metrics']
        for attr in exposed_attrs:
            setattr(self, attr, getattr(self._tuner, attr))
        if expose_fastai:
            cls = self.__class__
            self.__class__ = cls.__class__("AdaptiveTuner", (cls, _AdaptiveLearner), kwargs)

    def tune(
        self,
        epochs:int, # Number of iterations to train for
        lr:float = None, # If None, finds a new learning rate and uses suggestion_method
        strategy:Strategy = Strategy.OneCycle, # A fitting method
        callbacks:list = [], # Extra fastai Callbacks
        **kwargs ## kwargs for the fit function
    ):
        "Fine tune `self.model` for `epochs` with an `lr` and `strategy`"
        func = getattr(self, strategy, getattr(self._tuner, strategy, None))
        for attr in 'epochs,lr,cbs'.split():
            if attr in kwargs.keys(): kwargs.pop(attr)
        func(epochs, lr, cbs=callbacks, **kwargs)

    @delegates(Learner.lr_find)
    def lr_find(
        self,
        **kwargs # Learner.lr_find kwargs
    ):
        "Runs fastai's `LR Finder`"
        return self._tuner.lr_find(**kwargs)

    def save(
        self,
        save_directory # A folder to save our model to
    ):
        "Save a pretrained model to a `save_directory`"
        if rank_distrib(): return # Don't save if child proc
        self.model.save_pretrained(save_directory)
        self.tokenizer.save_pretrained(save_directory)
        return save_directory

    def load(
        self,
        path:Union[Path,str], # A location to load a tokenizer and weights from
        device = None # A valid device such as `cpu` or `cuda:0`
    ):
        "Loads a pretrained model with `AutoModel.from_pretrained` from `path` and loads it to device"
        if device is None and hasattr(self.dls, 'device'): device = self.dls.device
        self.model = AutoModel.from_pretrained(path)
        self.model = self.model.to(device)
        self.tokenizer = AutoTokenizer.from_pretrained(path)
        return self

    def predict(
        self,
        text:Union[List[str], str] # Some text or list of texts to inference with
    ):
        "Predict some `text` with the current model. Needs to be implemented for each task separately"
        raise NotImplementedError()

    def export(
        self,
        save_directory:Union[Path,str] # A folder to export our model to
    ):
        "Exports the current model and tokenizer information to `save_directory`"
        raise NotImplementedError()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_model.ipynb (unless otherwise specified).

__all__ = ['CudaCallback', 'AdaptiveModel']

# Cell
from typing import Union, List
from pathlib import Path
from abc import ABC, abstractmethod

from flair.data import Sentence

import torch
from torch import nn
from torch.utils.data import TensorDataset, DataLoader

from fastcore.basics import noop, store_attr, patch, first, ifnone
from fastcore.meta import delegates

from fastai.callback.core import Callback, GatherPredsCallback
from fastai.callback.progress import ProgressCallback

from fastai.learner import Learner
from fastai.data.core import DataLoaders

from fastai.torch_core import to_device, default_device

from .callback import GatherInputsCallback, SetInputsCallback

# Cell
@patch
def one_batch(
    self:DataLoader
) -> torch.Tensor: # A batch of data
    "Pathed functionality that grabs one batch of data from the `DataLoader` and deletes the iter"
    res = first(self)
    if hasattr(self, 'it'): delattr(self, 'it')
    return res

# Cell
@patch
def after_validate(
    self:GatherPredsCallback
):
    "Patched functionality that does nothing"
    pass

# Cell
class CudaCallback(Callback):
    "Move data to CUDA device"
    def __init__(
        self,
        device:str=None # A device to move the data to, such as 'cuda:0' or 'cpu'
    ): self.device = ifnone(device, default_device())
    def before_batch(self): self.learn.xb,self.learn.yb = to_device(self.xb),to_device(self.yb)
    def before_fit(self): self.model.to(self.device)

# Internal Cell
class _NoopModel(nn.Module):
    """
    Very basic model that performs y = mx+b. Is used only as a placeholder model for `BaseLearner`.
    Is based on https://github.com/fastai/fastai/blob/master/fastai/test_utils.py#L30
    """
    def __init__(self):
        super().__init__()
        self.a,self.b = nn.Parameter(torch.randn(1)),nn.Parameter(torch.randn(1))

    def forward(self, x): return x*self.a + self.b

# Internal Cell
class _BaseLearner:
    """
    Simple `Learner` class with `synth` DataLoaders, a noop model, and a noop loss function.

    Contains access to minimal `Learner` functionality including:
      - `get_preds`
      - `lr_find`, `fit_one_cycle`, `fit_flat_cos`, `fit_sgdr`, `fit` (not implemented)
      - `metrics`, `opt_func`, `splitter`, `wd`, `moms` (not implemented)
    """
    __cbs = [SetInputsCallback(), GatherInputsCallback()]

    def __init__(self, device='cuda' if torch.cuda.is_available() else 'cpu') -> None:
        """
        Generates blank `Learner` and stores it away privately.
        """
        self.__cbs.append(CudaCallback(device))
        self.__learner = Learner(self._generate_dls(), _NoopModel(), loss_func=noop, cbs=self.__cbs)
        self.__default_dls, self.__default_model = True, True

    def _generate_dls(self, a=2, b=3, batch_size=16, n_train=10, n_valid=2) -> DataLoaders:
        """
        Builds synthetic `DataLoaders`.
        Implementation based on https://github.com/fastai/fastai/blob/master/fastai/test_utils.py#L18
        """
        def get_data(n) -> TensorDataset:
            """
            Generates synthetic `TensorDataset`
            """
            x = torch.randn(batch_size*n, 1)
            return TensorDataset(x, a*x + b, 0.1*torch.randn(batch_size*n, 1))

        train_ds = get_data(n_train)
        valid_ds = get_data(n_valid)
        train_dl = DataLoader(train_ds, batch_size=batch_size, shuffle=True, num_workers=0)
        valid_dl = DataLoader(valid_ds, batch_size=batch_size, num_workers=0)
        return DataLoaders(train_dl, valid_dl)

    def get_preds(self, dl=None, cbs=[]):
        """
        Get raw predictions based on `dl` with `cbs`.

        For basic inference, `cbs` should include any `Callbacks` needed to do general inference
        """
        if dl is None: raise ValueError("`dl` should not be `None`")
        if isinstance(self.__learner.model, _NoopModel):
            raise ValueError("The default model is still set, you should override this with `_BaseLearner.set_model(x)`")
        return self.__learner.get_preds(dl=dl, cbs=cbs)

    def set_device(self, device:str='cpu'):
        if device != 'cpu' and device != 'cuda':
            raise ValueError("Device must either be `cpu` or `cuda`")
        self.__learn.__cbs[-1].device = device

    def set_as_dict(self, as_dict:bool=False):
        """
        Sets `is_dict` in the `SetInputsCallback`. Should be utilized whenever dictating
        if the incoming batch should be a dictionary or the actual inputs.
        """
        for cb in self.__cbs:
            if isinstance(cb, SetInputsCallback):
                cb.as_dict = as_dict

    def set_model(self, model):
        """
        Sets `Learner`'s model to `model`
        """
        self.__learner.model = model

# Cell
class AdaptiveModel(ABC):
    _learn = _BaseLearner()
    def set_model(
        self,
        model # A PyTorch model
    ):
        "Sets model in `_learn`"
        self._learn.set_model(model)
        self.model = model

    def set_as_dict(
        self,
        as_dict:bool=False # Whether to return the inputs as a dictionary when predicting or training
    ):
        "Sets `as_dict` in `_learn`"
        self._learn.set_as_dict(as_dict)

    def set_device(
        self,
        device:str='cpu' #  A device for the `CudaCallback`, such as 'cuda:0' or 'cpu'
    ):
        "Sets the device for `CudaCallback` in `__learn`"
        self._learn.set_device(device)


    def get_preds(
        self,
        dl=None, # An iterable DataLoader or DataLoader-like object
        cbs=[] # Optional fastai `Callbacks`
    ):
        """
        Get raw predictions based on `dl` with `cbs`.

        For basic inference, `cbs` should include any `Callbacks` needed to do general inference
        """
        return self._learn.get_preds(dl=dl, cbs=cbs)

    @abstractmethod
    def load(
        self,
        model_name_or_path: Union[str, Path], # A model file location to use
    ):
        "Load model into the `AdaptiveModel` object as alternative constructor"
        raise NotImplementedError("Please Implement this method")

    @abstractmethod
    def predict(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Some text to predict on
        mini_batch_size: int = 32, # A batch size for if a list of texts are passed in
        **kwargs,
    ) -> List[Sentence]: # A list of predicted sentences
        "Run inference on the model"
        raise NotImplementedError("Please Implement this method")


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_embeddings.ipynb (unless otherwise specified).

__all__ = ['logger', 'EmbeddingResult', 'EasyWordEmbeddings', 'EasyStackedEmbeddings', 'EasyDocumentEmbeddings']

# Cell
import logging, torch
from typing import List, Dict, Union
from fastcore.basics import listify
from collections import defaultdict, OrderedDict

from fastcore.basics import mk_class
from fastcore.xtras import dict2obj
from fastcore.dispatch import typedispatch
from flair.data import Sentence
from flair.embeddings import (
    Embeddings,
    WordEmbeddings,
    StackedEmbeddings,
    FlairEmbeddings,
    DocumentPoolEmbeddings,
    DocumentRNNEmbeddings,
    TransformerWordEmbeddings,
)

from ..model_hub import FlairModelResult, HFModelResult, HFModelHub, FlairModelHub

from ..result import SentenceResult, DetailLevel

# Cell
_flair_hub = FlairModelHub()
_hf_hub = HFModelHub()

# Cell
logger = logging.getLogger(__name__)

# Internal Cell
@typedispatch
def _make_sentences(text:str, as_list=False) -> Union[List[Sentence], Sentence]:
    return [Sentence(text)] if as_list else Sentence(text)

# Internal Cell
@typedispatch
def _make_sentences(text:list, as_list=False) -> Union[List[Sentence], Sentence]:
    if all(isinstance(t,str) for t in text):
        return [Sentence(t) for t in text]
    elif all(isinstance(t, Sentence) for t in text):
        return text

# Internal Cell
@typedispatch
def _make_sentences(text:Sentence, as_list=False) -> Union[List[Sentence], Sentence]:
    return [text] if as_list else text

# Internal Cell
@typedispatch
def _get_embedding_model(model_name_or_path:HFModelResult) -> TransformerWordEmbeddings:
    return TransformerWordEmbeddings(model_name_or_path.name)

# Internal Cell
@typedispatch
def _get_embedding_model(model_name_or_path:FlairModelResult) -> Union[FlairEmbeddings, WordEmbeddings]:
    nm = model_name_or_path.name
    try:
        return WordEmbeddings(nm.strip('flairNLP/'))
    except:
        return FlairEmbeddings(nm.strip('flairNLP/'))

# Internal Cell
@typedispatch
def _get_embedding_model(model_name_or_path:str) -> Union[TransformerWordEmbeddings, WordEmbeddings, FlairEmbeddings]:
    res = _flair_hub.search_model_by_name(model_name_or_path, user_uploaded=True)
    if len(res) < 1:
        # No models found
        res = _hf_hub.search_model_by_name(model_name_or_path, user_uploaded=True)
        if len(res) < 1:
            raise ValueError(f'Embeddings not found for the model key: {model_name_or_path}, check documentation or custom model path to verify specified model')
        else:
            return TransformerWordEmbeddings(res[0].name) # Returning the first should always be the non-fast option
    else:
        nm = res[0].name
        try:
            return WordEmbeddings(nm.strip('flairNLP/'))
        except:
            return FlairEmbeddings(nm.strip('flairNLP/'))

# Cell
class EmbeddingResult(SentenceResult):
    "A result class designed for Embedding models"
    def __init__(
        self,
        sentences:List[Sentence] # A list of Flair `Sentence`s
    ): super().__init__(sentences)

    @property
    def sentence_embeddings(self) -> List[torch.tensor]:
        "All embeddings in `sentences` (if available)"
        return [s.get_embedding() for s in self._sentences]

    @property
    def token_embeddings(self) -> List[torch.tensor]:
        "All embeddings from the individual tokens in `sentence` with original order in shape (n, embed_dim)"
        return [torch.stack([tok.get_embedding() for tok in s], dim=0) for s in self._sentences]

    def to_dict(
        self,
        detail_level:DetailLevel=DetailLevel.Low # A level of detail to return
    ):
        "Returns `self` as a dictionary"
        o = OrderedDict()
        o.update({'inputs':self.inputs,
                  'sentence_embeddings':self.sentence_embeddings,
                 'token_embeddings':self.token_embeddings})
        if detail_level == "low": return o
        if detail_level == 'medium' or detail_level == 'high':
            # Return embeddings/word pairs and indicies, and the tokenized input
            for s in self._sentences:
                o.update({
                    tok.text:{
                        'embeddings':tok.get_embedding(),
                        'word_idx':tok.idx
                    } for tok in s
                })
            o.update({
                'tokenized_inputs':self.tokenized_inputs
            })
        if detail_level == 'high':
            # Return embeddings/word pairs, indicies, and the original Sentences objects
            for s in self._sentences:
                o.update({tok.text:{
                    'embeddings':tok.get_embedding(),
                    'word_idx':tok.idx
                    } for tok in s})
            o.update({'sentences':self._sentences})
        return o

    def __repr__(self):
        s = f"{self.__class__.__name__}:" + " {"
        s += f'\n\tInputs: {self.inputs}'
        if self.token_embeddings is not None: s += f'\n\tToken Embeddings Shapes: {[f.shape for f in self.token_embeddings]}'
        if self.sentence_embeddings is not None: s += f'\n\tSentence Embeddings Shapes: {[f.shape for f in self.sentence_embeddings]}'
        return s + '\n}'

# Cell
class EasyWordEmbeddings:
    """Word embeddings from the latest language models

    Usage:

    ```python
    >>> embeddings = adaptnlp.EasyWordEmbeddings()
    >>> embeddings.embed_text("text you want embeddings for", model_name_or_path="bert-base-cased")
    ```
    """

    def __init__(self):
        self.models: Dict[Embeddings] = defaultdict(bool)

    def embed_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        model_name_or_path: Union[str, HFModelResult, FlairModelResult] = "bert-base-cased", # The hosted model name key, model path, or an instance of either `HFModelResult` or `FlairModelResult`
        detail_level:DetailLevel = DetailLevel.Low, # A level of detail to return. By default is None, which returns a EmbeddingResult, otherwise will return a dict
        raw:bool=False # Whether to return the raw outputs
    ) -> List[EmbeddingResult]: # A list of either `EmbeddingResult`s or dictionaries with information
        "Produces embeddings for text"
        # Convert into sentences
        sentences = _make_sentences(text)

        # Load correct Embeddings module
        if model_name_or_path not in self.models.keys():
            self.models[model_name_or_path] = _get_embedding_model(model_name_or_path)
        embedding = self.models[model_name_or_path]
        embeds = embedding.embed(sentences)

        if not raw:
            res = EmbeddingResult(listify(embeds))
            return res.to_dict(detail_level) if detail_level is not None else res
        else:
            return listify(embeds)

    def embed_all(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        model_names_or_paths:List[str] = [], # A list of model names
        detail_level:DetailLevel=DetailLevel.Low, # A level of detail to return. By default is None, which returns a EmbeddingResult, otherwise will return a dict
    ) -> List[EmbeddingResult]: # A list of either `EmbeddingResult`s or dictionaries with information
        "Embeds text with all embedding models loaded"
        # Convert into sentences
        sentences = _make_sentences(text)

        if model_names_or_paths:
            for embedding_name in model_names_or_paths:
                sentences = self.embed_text(
                    sentences, model_name_or_path=embedding_name, raw=True
                )
        else:
            for embedding_name in self.models.keys():
                sentences = self.embed_text(
                    sentences, model_name_or_path=embedding_name, raw=True
                )
        res = EmbeddingResult(listify(sentences))
        return res.to_dict(detail_level) if detail_level is not None else res

# Cell
class EasyStackedEmbeddings:
    "Word Embeddings that have been concatenated and 'stacked' as specified by Flair"
    def __init__(
        self,
        *embeddings: str
    ):
        print("May need a couple moments to instantiate...")
        self.embedding_stack = []

        # Load correct Embeddings module
        for model_name_or_path in embeddings:
            self.embedding_stack.append(_get_embedding_model(model_name_or_path))

        assert len(self.embedding_stack) != 0
        self.stacked_embeddings = StackedEmbeddings(embeddings=self.embedding_stack)

    def embed_text(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        detail_level:DetailLevel = DetailLevel.Low # A level of detail to return. By default is None, which returns a EmbeddingResult, otherwise will return a dict
    ) -> List[EmbeddingResult]: # A list of either EmbeddingResult's or dictionaries with information
        "Stacked embeddings"
        # Convert into sentences
        sentences = _make_sentences(text, as_list=True)

        # Unlike flair embeddings modules, stacked embeddings do not return a list of sentences
        self.stacked_embeddings.embed(sentences)

        res = EmbeddingResult(listify(sentences))
        return res.to_dict(detail_level) if detail_level is not None else res

# Cell
class EasyDocumentEmbeddings:
    "Document Embeddings generated by pool and rnn methods applied to the word embeddings of text"

    __allowed_methods = ["rnn", "pool"]
    __allowed_configs = ("pool_configs", "rnn_configs")

    def __init__(
        self,
        *embeddings: str, # Variable number of strings referring to model names or paths
        methods: List[str] = ["rnn", "pool"], # A list of strings to specify which document embeddings to use i.e. ["rnn", "pool"] (avoids unncessary loading of models if only using one)
        configs: Dict = {
            "pool_configs": {"fine_tune_mode": "linear", "pooling": "mean"},
            "rnn_configs": {
                "hidden_size": 512,
                "rnn_layers": 1,
                "reproject_words": True,
                "reproject_words_dimension": 256,
                "bidirectional": False,
                "dropout": 0.5,
                "word_dropout": 0.0,
                "locked_dropout": 0.0,
                "rnn_type": "GRU",
                "fine_tune": True,
            },
        }, # A dictionary of configurations for flair's rnn and pool document embeddings
    ):
        print("May need a couple moments to instantiate...")
        self.embedding_stack = []

        # Check methods
        for m in methods:
            assert m in self.__class__.__allowed_methods

        # Set configs for pooling and rnn parameters
        for k, v in configs.items():
            assert k in self.__class__.__allowed_configs
            setattr(self, k, v)

        # Load correct Embeddings module
        for model_name_or_path in embeddings:
            self.embedding_stack.append(_get_embedding_model(model_name_or_path))

        assert len(self.embedding_stack) != 0
        if "pool" in methods:
            self.pool_embeddings = DocumentPoolEmbeddings(
                self.embedding_stack, **self.pool_configs
            )
            print("Pooled embedding loaded")
        if "rnn" in methods:
            self.rnn_embeddings = DocumentRNNEmbeddings(
                self.embedding_stack, **self.rnn_configs
            )
            print("RNN embeddings loaded")

    def embed_pool(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        detail_level:DetailLevel = DetailLevel.Low, # A level of detail to return. By default is None, which returns a EmbeddingResult, otherwise will return a dict
    ) -> List[EmbeddingResult]: # A list of either EmbeddingResult's or dictionaries with information
        "Generate stacked embeddings with `DocumentPoolEmbeddings`"
        sentences = _make_sentences(text, as_list=True)
        self.pool_embeddings.embed(sentences)
        res = EmbeddingResult(listify(sentences))
        return res.to_dict(detail_level) if detail_level is not None else res

    def embed_rnn(
        self,
        text: Union[List[Sentence], Sentence, List[str], str], # Text input, it can be a string or any of Flair's `Sentence` input formats
        detail_level:DetailLevel = DetailLevel.Low, # A level of detail to return. By default is None, which returns a EmbeddingResult, otherwise will return a dict
    ) -> List[Sentence]: # A list of either EmbeddingResult's or dictionaries with information
        "Generate stacked embeddings with `DocumentRNNEmbeddings`"
        sentences = _make_sentences(text, as_list=True)
        self.rnn_embeddings.embed(sentences)
        res = EmbeddingResult(listify(sentences))
        return res.to_dict(detail_level) if detail_level is not None else res


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/14_training.language_model.ipynb (unless otherwise specified).

__all__ = ['LanguageModelDatasets', 'LanguageModelTuner']

# Cell
import pandas as pd
from fastcore.basics import mk_class
from fastcore.meta import delegates
from fastcore.xtras import Path, range_of

from fastai.basics import *
from fastai.data.core import DataLoaders
from fastai.data.transforms import get_files

from .core import *
from .arrow_utils import TextNoNewLineDatasetReader

from ..inference.text_generation import TransformersTextGenerator

from transformers import DataCollatorForLanguageModeling, default_data_collator
from datasets import Dataset

from typing import List

# Internal Cell
def _group_texts(examples, block_size):
    # Concatenate all texts, based on code by Transformers
    concatenated_examples = {k: sum(examples[k], []) for k in examples.keys()}
    total_length = len(concatenated_examples[list(examples.keys())[0]])
    # We drop the small remainder, we could add padding if the model supported it instead of this drop, you can
    # customize this part to your needs.
    total_length = (total_length // block_size) * block_size
    # Split by chunks of max_len.
    result = {
        k: [t[i : i + block_size] for i in range(0, total_length, block_size)]
        for k, t in concatenated_examples.items()
    }
    result["labels"] = result["input_ids"].copy()
    return result

# Cell
def _tokenize(item, tokenizer, tokenize_kwargs): return tokenizer(item['text'], **tokenize_kwargs)

# Cell
class LanguageModelDatasets(TaskDatasets):
    "A set of datasets designed for language model fine-tuning"
    def __init__(
        self,
        train_dset:Dataset, # A training dataset
        valid_dset:Dataset, # A validation dataset
        tokenizer_name:str, # The name of a tokenizer
        tokenize:bool, # Whether to tokenize immediatly
        tokenize_kwargs:dict, # kwargs for the tokenize function
        auto_kwargs:dict, # AutoTokenizer.from_pretrained kwargs
        remove_columns:list, # The columns to remove when tokenizing
        block_size:int, # The size of each block
        masked_lm:bool # Whether the language model is a MLM
    ):
        super().__init__(
            train_dset,
            valid_dset,
            tokenizer_name,
            tokenize,
            _tokenize,
            tokenize_kwargs,
            auto_kwargs
        )
        self.masked_lm = masked_lm
        self.block_size = block_size
        f = partial(_group_texts, block_size=self.block_size)
        self.train = self.train.map(f, batched=True)
        self.valid = self.valid.map(f, batched=True)

    @classmethod
    def from_dfs(
        cls,
        train_df:pd.DataFrame, # A Pandas Dataframe
        text_col:str, # The name of the text column
        tokenizer_name:str, # The name of the tokenizer
        block_size:int=128, # The size of each block
        masked_lm:bool=False, # Whether the language model is a MLM
        valid_df:pd.DataFrame=None, # An optional validation DataFrame
        split_func:callable=None, # Optionally a splitting function similar to RandomSplitter
        split_pct:float=.1, # What % to split the df between training and validation
        tokenize_kwargs:dict={}, # kwargs for the tokenize function
        auto_kwargs:dict={} # kwargs for the AutoTokenizer.from_pretrained constructor
    ):
        "Builds `LanguageModelDatasets` from a `DataFrame` or file path"
        if split_func is None: split_func = RandomSplitter(split_pct)
        if valid_df is None:
            train_idxs, valid_idxs = split_func(range_of(train_df))
            valid_df = train_df.iloc[valid_idxs]
            train_df = train_df.iloc[train_idxs]

        train_df = train_df[[text_col]]
        valid_df = valid_df[[text_col]]

        train_df = train_df.rename(columns={text_col:'text'})
        valid_df = valid_df.rename(columns={text_col:'text'})

        train_dset = Dataset.from_dict(train_df.to_dict('list'))
        valid_dset = Dataset.from_dict(valid_df.to_dict('list'))

        dsets = TaskDatasets(
            train_dset,
            valid_dset,
            tokenizer_name,
            False,
            _tokenize,
            tokenize_kwargs=tokenize_kwargs,
            auto_kwargs=auto_kwargs
        )

        f = partial(_group_texts, block_size=512)
        t = partial(_tokenize, tokenizer=dsets.tokenizer, tokenize_kwargs=tokenize_kwargs)
        dsets.train = dsets.train.map(t, batched=True, remove_columns=['text'])
        dsets.valid = dsets.valid.map(t, batched=True, remove_columns=['text'])

        dsets.train = dsets.train.map(f, batched=True)
        dsets.valid = dsets.valid.map(f, batched=True)
        return dsets

    @classmethod
    def from_csvs(
        cls,
        train_csv:Path, # A training csv file
        text_col:str, # The name of the text column
        tokenizer_name:str, # The name of the tokenizer
        block_size:int=128, # The size of each block
        masked_lm:bool=False, # Whether the language model is a MLM
        valid_csv:Path=None, # An optional validation csv
        split_func:callable=None, # Optionally a splitting function similar to RandomSplitter
        split_pct:float=.1, # What % to split the df between training and validation
        tokenize_kwargs:dict={}, # kwargs for the tokenize function
        auto_kwargs:dict={}, # kwargs for the AutoTokenizer.from_pretrained constructor
        **kwargs, # kwargs for `pd.read_csv`
    ):
        "Builds `LanguageModelDatasets` from a single csv or set of csvs. A convience constructor for `from_dfs`"
        train_df = pd.read_csv(train_csv, **kwargs)
        if valid_csv is not None: valid_df = pd.read_csv(valid_csv, **kwargs)
        else: valid_df = None
        return cls.from_dfs(
            train_df,
            text_col,
            tokenizer_name,
            block_size,
            masked_lm,
            valid_df,
            split_func,
            split_pct,
            tokenize_kwargs,
            auto_kwargs
        )

    @classmethod
    def from_folders(
        cls,
        train_path:Path, # The path to the training data
        tokenizer_name:str, # The name of the tokenizer
        block_size:int=128, # The size of each block
        masked_lm:bool=False, # Whether the language model is a MLM
        valid_path:Path=None, # An optional validation path
        split_func:callable=None, # Optionally a splitting function similar to RandomSplitter
        split_pct:float=.1, # What % to split the df between training and validation
        tokenize_kwargs:dict={}, # kwargs for the tokenize function
        auto_kwargs:dict={} # kwargs for the AutoTokenizer.from_pretrained constructor
    ):
        "Builds `LanguageModelDatasets` from a folder or group of folders"
        train_txts = get_files(train_path, extensions='.txt')
        if valid_path is not None:
            valid_txts = get_files(valid_path, extensions='.txt')
        else:
            if split_func is not None:
                split_func = RandomSplitter(split_pct)
            train_idxs, valid_idxs = split_func(train_txts)
            valid_txts = train_txts[valid_idxs]
            train_txts = train_txts[train_idxs]
        train_txts = [str(x) for x in train_txts]
        valid_txts = [str(x) for x in valid_txts]
        train_dset = TextNoNewLineDatasetReader(train_txts).read()
        valid_dset = TextNoNewLineDatasetReader(valid_txts).read()

        dsets = TaskDatasets(
            train_dset,
            valid_dset,
            tokenizer_name,
            False,
            _tokenize,
            tokenize_kwargs=tokenize_kwargs,
            auto_kwargs=auto_kwargs
        )

        f = partial(_group_texts, block_size=512)
        t = partial(_tokenize, tokenizer=dsets.tokenizer, tokenize_kwargs=tokenize_kwargs)
        dsets.train = dsets.train.map(t, batched=True, remove_columns=['text'])
        dsets.valid = dsets.valid.map(t, batched=True, remove_columns=['text'])

        dsets.train = dsets.train.map(f, batched=True)
        dsets.valid = dsets.valid.map(f, batched=True)
        return dsets

    @delegates(DataLoaders)
    def dataloaders(
        self,
        batch_size=8, # A batch size
        shuffle_train=True, # Whether to shuffle the training dataset
        collate_fn = default_data_collator, # A custom collation function
        mlm_probability:float = 0.15, # Token masking probablity for Masked Language Models
        **kwargs
    ):
        "Build DataLoaders from `self`"
        if self.masked_lm: collate_fn = DataCollatorForLanguageModeling(tokenizer=self.tokenizer, mlm_probability=mlm_probability)
        return super().dataloaders(batch_size, shuffle_train, collate_fn, **kwargs)

# Cell
from transformers import AutoModelForMaskedLM, AutoModelForCausalLM, AutoModelForSeq2SeqLM

mk_class('LMType', **{o:o.lower() for o in ['Masked', 'Causal', 'Seq2Seq']},
        doc="All valid language model classes with typo-proofing")

_constructors = {
    'masked':AutoModelForMaskedLM.from_pretrained,
    'causal':AutoModelForCausalLM.from_pretrained,
    'seq2seq':AutoModelForSeq2SeqLM.from_pretrained
}

# Cell
class LanguageModelTuner(AdaptiveTuner):
    """
    An `AdaptiveTuner` with good defaults for Language Model fine-tuning
    **Valid kwargs and defaults:**
      - `lr`:float = 0.001
      - `splitter`:function = `trainable_params`
      - `cbs`:list = None
      - `path`:Path = None
      - `model_dir`:Path = 'models'
      - `wd`:float = None
      - `wd_bn_bias`:bool = False
      - `train_bn`:bool = True
      - `moms`: tuple(float) = (0.95, 0.85, 0.95)
    """
    def __init__(
        self,
        dls:DataLoaders, # A set of DataLoaders or AdaptiveDataLoaders
        model_name, # A HuggingFace model
        tokenizer = None, # A HuggingFace tokenizer
        language_model_type:LMType = LMType.Causal, # The type of language model to use
        loss_func = CrossEntropyLossFlat(), # A loss function
        metrics = [Perplexity()], # Metrics to monitor the training with
        opt_func = Adam, # A fastai or torch Optimizer
        additional_cbs = None, # Additional Callbacks to have always tied to the Tuner,
        expose_fastai_api = False, # Whether to expose the fastai API
        **kwargs, # kwargs for `Learner.__init__`
    ):
        additional_cbs = listify(additional_cbs)
        for arg in 'dls,model,loss_func,metrics,opt_func,cbs,expose_fastai'.split(','):
            if arg in kwargs.keys(): kwargs.pop(arg) # Pop all existing kwargs

        if language_model_type is None: raise ValueError("Please specify the type of language model you want to use, such as `masked` or `causal`")
        if language_model_type not in _constructors.keys():
            raise ValueError(
                """
                Please enter a valid Langauge Model Type of:
                  * `masked` or `LMType.Masked`
                  * `causal` or `LMType.Causal`
                  * `seq2seq` or `LMType.Seq2Seq`
                """
            )
        try:
            model = _constructors[language_model_type](model_name)
        except Exception as e:
            message = e.args[0]
            m = f"Was not able to create a {language_model_type} instance of {model_name}. Please use a valid model for your task:"
            m += message
            e.args = [m]
            raise e

        if tokenizer is None: tokenizer = AutoTokenizer.from_pretrained(model_name)
        tokenizer.pad_token = "<PAD>"

        super().__init__(
            expose_fastai_api,
            dls = dls,
            model = model,
            tokenizer = tokenizer,
            loss_func = loss_func,
            metrics = metrics,
            opt_func = opt_func,
            cbs=additional_cbs,
            **kwargs
        )

    def predict(
        self,
        text:Union[List[str], str], # Some text or list of texts to do inference with
        bs:int=64, # A batch size to use for multiple texts
        num_tokens_to_produce:int=50, # Number of tokens to generate
        **kwargs, # Optional arguments for `PretrainedModel.generate`
    ):
        "Predict some `text` for sequence classification with the currently loaded model"
        if getattr(self, '_inferencer', None) is None:

            if self.tokenizer.pad_token is None:
                self.tokenizer.pad_token="<PAD>"
            self._inferencer = TransformersTextGenerator(self.tokenizer,self.model)
        return self._inferencer.predict(text, bs, num_tokens_to_produce, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/15_training.arrow_utils.ipynb (unless otherwise specified).

__all__ = ['TextNoNewLine', 'TextNoNewLineDatasetReader']

# Cell
from typing import Optional

from datasets import Features, NamedSplit
from datasets.packaged_modules.text.text import Text
from datasets.utils.typing import NestedDataStructureLike, PathLike
from datasets.io.abc import AbstractDatasetReader
import datasets
from datasets.packaged_modules.text.text import TextConfig

import pyarrow as pa

# Cell
class TextNoNewLine(datasets.ArrowBasedBuilder):
    "An ArrowBasedBuilder dataset that doesn't split based on newlines"
    BUILDER_CONFIG_CLASS = TextConfig

    def _info(self):
        return datasets.DatasetInfo(features=self.config.features)

    def _split_generators(self, dl_manager):
        """The `data_files` kwarg in load_dataset() can be a str, List[str], Dict[str,str], or Dict[str,List[str]].
        If str or List[str], then the dataset returns only the 'train' split.
        If dict, then keys should be from the `datasets.Split` enum.
        """
        if not self.config.data_files:
            raise ValueError(f"At least one data file must be specified, but got data_files={self.config.data_files}")
        data_files = dl_manager.download_and_extract(self.config.data_files)
        if isinstance(data_files, (str, list, tuple)):
            files = data_files
            if isinstance(files, str):
                files = [files]
            return [datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={"files": files})]
        splits = []
        for split_name, files in data_files.items():
            if isinstance(files, str):
                files = [files]
            splits.append(datasets.SplitGenerator(name=split_name, gen_kwargs={"files": files}))
        return splits

    def _generate_tables(self, files):
        schema = pa.schema(self.config.features.type if self.config.features is not None else {"text": pa.string()})
        for file_idx, file in enumerate(files):
            batch_idx = 0
            with open(file, "r", encoding=self.config.encoding) as f:
                while True:
                    batch = f.read(self.config.chunksize)
                    if not batch:
                        break
                    batch = [batch]
                    pa_table = pa.Table.from_arrays([pa.array(batch)], schema=schema)
                    # Uncomment for debugging (will print the Arrow table size and elements)
                    # logger.warning(f"pa_table: {pa_table} num rows: {pa_table.num_rows}")
                    # logger.warning('\n'.join(str(pa_table.slice(i, 1).to_pydict()) for i in range(pa_table.num_rows)))
                    yield (file_idx, batch_idx), pa_table
                    batch_idx += 1

# Cell
class TextNoNewLineDatasetReader(AbstractDatasetReader):
    "A `DatasetReader` class that mimics `TextDatasetReader`, but uses `TextNoNewLine`"
    def __init__(
        self,
        path_or_paths: NestedDataStructureLike[PathLike], # Paths to texts
        split: Optional[NamedSplit] = None, # Potential dataset splits
        features: Optional[Features] = None, # Any features for the datasets
        cache_dir: str = None, # A cache directory
        keep_in_memory: bool = False, # Whether to keep the dataset completely in memory
        **kwargs,
    ):
        super().__init__(
            path_or_paths, split=split, features=features, cache_dir=cache_dir, keep_in_memory=keep_in_memory, **kwargs
        )
        path_or_paths = path_or_paths if isinstance(path_or_paths, dict) else {self.split: path_or_paths}
        self.builder = TextNoNewLine(
            cache_dir=cache_dir,
            data_files=path_or_paths,
            features=features,
            **kwargs,
        )

    def read(self):
        "Reads in a dataset"
        download_config = None # A download configuration
        download_mode = None # A download mode
        ignore_verifications = False # Whether to ignore any verification requests
        use_auth_token = None # Potential authorization tokens
        base_path = None # A path

        self.builder.download_and_prepare(
            download_config=download_config,
            download_mode=download_mode,
            ignore_verifications=ignore_verifications,
            # try_from_hf_gcs=try_from_hf_gcs,
            base_path=base_path,
            use_auth_token=use_auth_token,
        )

        # Build dataset for splits
        dataset = self.builder.as_dataset(
            split=self.split, ignore_verifications=ignore_verifications, in_memory=self.keep_in_memory
        )
        return dataset